/*----------------------------------------------------------------------
  SFML Benchmark - Roberto Prieto 
  Copyright (C) 2013-2016 MegaStorm Systems
  contact@megastormsystems.com - http://www.megastormsystems.com

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

 ------------------------------------------------------------------------
                                SFML Benchmark Info
    
    - Uses SFML 2.3.2
    - OpenGL
    - 32 version

    Changelog:
 
 v1.10 Mar/2016 - Uses SFML 2.3.2.
 v1.00 Sept/2015 - First version.
   
 ------------------------------------------------------------------------ */


#include <windows.h>
#include <time.h>
#include "SFML/Graphics.hpp"
#include "../SDL-Benchmark.h"

// Common vars
typedef struct SFML_Rect
{
    int x,y;
    int w,h;
} SFML_Rect;
static int sprite_w, sprite_h;
static SFML_Rect *positions;
static SFML_Rect *velocities;
SDLBenchParameters SDLBenchParams; // Parameters struct

// SFML vars
sf::RenderWindow *window;
sf::Texture **pTextureTable = NULL;
sf::Sprite **pSpriteTable = NULL;
//#define DEBUG_RENDERING

int RemoveSprites(int iNumRealSpr)
{
    int i = 0;

    while(i < iNumRealSpr)
    {
        if(pSpriteTable != NULL)
        {
            if(pSpriteTable[i] != NULL) delete(pSpriteTable[i]);
            pSpriteTable[i] = NULL;
        }
        if(pTextureTable != NULL)
        {
            if(pTextureTable[i] != NULL) delete(pTextureTable[i]);
            pTextureTable[i] = NULL;
        }
        ++i;        
    }
    free(pSpriteTable);    
    pSpriteTable = NULL;
    free(pTextureTable);
    pTextureTable = NULL;
    return 0;
}

int LoadSprites(char *file, int iNumRealSpr)
{
    int i = 0;   
    sf::Image imgTmp;
    sf::Color colorBlack;    

    // Load the image
    if(!imgTmp.loadFromFile(SDLBenchParams.szIcon)) return -1;
    colorBlack.r = 0; colorBlack.g = 0; colorBlack.b = 0;
    imgTmp.createMaskFromColor(colorBlack);

    // Create textures for each sprite using the surface
    int bFlagError = 0;
    pSpriteTable = new sf::Sprite *[iNumRealSpr];    
    pTextureTable = new sf::Texture *[iNumRealSpr];
    while(i < iNumRealSpr)
    {      
        pTextureTable[i] = new sf::Texture;
        if(!pTextureTable[i]->loadFromImage(imgTmp)) 
        {
            fprintf(stderr, "Couldn't create texture for sprite %d\n", i);
            bFlagError = 1;
        }
        else
        {   
            sf::Vector2u vSize= pTextureTable[i]->getSize();
            sprite_w = vSize.x;
            sprite_h = vSize.y;            
            pSpriteTable[i] = new sf::Sprite;
            pSpriteTable[i]->setTexture(*pTextureTable[i]);
        }
        ++i;        
    }
    
    if(bFlagError == 1) 
    {
        RemoveSprites(iNumRealSpr);
        return -2;
    }
    return 0;
}

void MoveSprites(int iNumSpr, int iNumRealSpr)
{
    int i, n;
    SFML_Rect *position, *velocity;
    #ifdef DEBUG_RENDERING
     LARGE_INTEGER StartingTime, EndingTime, measure1, measure2;
     QueryPerformanceCounter(&StartingTime);
    #endif

    // Grey background
    window->clear(sf::Color(0xA0, 0xA0, 0xA0));
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&measure1);
    #endif

    // Move the sprite, bounce at the wall, and draw
    n = 0;
    for(i = 0; i < iNumSpr; ++i) 
    {
        position = &positions[i];
        velocity = &velocities[i];
        position->x += velocity->x;
        if ((position->x < 0) || (position->x >= (SDLBenchParams.iVideoWidth - sprite_w))) 
        {
            velocity->x = -velocity->x;
            position->x += velocity->x;
        }
        position->y += velocity->y;
        if ((position->y < 0) || (position->y >= (SDLBenchParams.iVideoHeight - sprite_h))) 
        {
            velocity->y = -velocity->y;
            position->y += velocity->y;
        }

        // Render the sprite
        if(iNumRealSpr == 1) 
        {            
            pSpriteTable[0]->setPosition((float)position->x, (float)position->y);
            window->draw(*pSpriteTable[0]);
        }
        else
        {
            pSpriteTable[i]->setPosition((float)position->x, (float)position->y);
            window->draw(*pSpriteTable[i]);
        }
                
    }
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&measure2);
    #endif

    // Update the screen!
    window->display();

    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&EndingTime);
     printf("FrameTime: %llu + %llu + %llu = %llu\n", 
        measure1.QuadPart - StartingTime.QuadPart, 
        measure2.QuadPart - measure1.QuadPart, 
        EndingTime.QuadPart - measure2.QuadPart, 
        EndingTime.QuadPart - StartingTime.QuadPart);     
    #endif
}

int main(int argc, char *argv[])
{
    int i, iMaxSprites, iRealSprites, done;
    int iStart, iFrames;
    double fFPS_1, fFPS_2;

    // Print information and argument parsing
    parseAndInformation(argc, argv, SDLBENCH_SFML_VERSION" Benchmark", &SDLBenchParams);

    // Max and real sprites
    iMaxSprites = SDLBenchParams.iNumSprites;
    if(iMaxSprites < SDLBenchParams.iNumSprites2) iMaxSprites = SDLBenchParams.iNumSprites2;
    if(SDLBenchParams.iFillRateMode == 0) iRealSprites = iMaxSprites;
    else iRealSprites = 1;

    // Initialize SFML
    window = new sf::RenderWindow(sf::VideoMode(SDLBenchParams.iVideoWidth, SDLBenchParams.iVideoHeight, 32), SDLBENCH_SFML_VERSION); //, sf::Style::Fullscreen);
    window->setVerticalSyncEnabled(false);
    window->setMouseCursorVisible(false);
    printf("SFML renderer: opengl\n");

    // Load the texture and create the sprite
    if(LoadSprites(SDLBenchParams.szIcon, iRealSprites) < 0) 
    {
        delete window;
        fprintf(stderr, "Couldn't load %s\n", SDLBenchParams.szIcon);
        return -2;
    }
       
    // Allocate memory for the sprite info and initialize it    
    positions = (SFML_Rect *) malloc(iMaxSprites * sizeof(SFML_Rect));
    velocities = (SFML_Rect *) malloc(iMaxSprites * sizeof(SFML_Rect));
   
    srand((unsigned int)time(NULL));
    for(i = 0; i < iMaxSprites; ++i) 
    {
        positions[i].x = rand() % (SDLBenchParams.iVideoWidth - sprite_w);
        positions[i].y = rand() % (SDLBenchParams.iVideoHeight - sprite_h);
        positions[i].w = sprite_w;
        positions[i].h = sprite_h;
        velocities[i].x = 0;
        velocities[i].y = 0;
        while(!velocities[i].x && !velocities[i].y) 
        {
            velocities[i].x = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
            velocities[i].y = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
        }
    }

    // ---Main render loop for iNumSprites---    
    iFrames = 0;	
    iStart = clock();	
    done = 0;	
    while(!done) 
    {				
        ++iFrames;       
        MoveSprites(SDLBenchParams.iNumSprites, iRealSprites);
        if(( (float)(clock() - iStart) / 1000) >= SDLBenchParams.dBaseTime) { done = 1; }
    }
    // Print out some timing information
    if(clock() > iStart) 
    {
        fFPS_1 = (double)iFrames / (double)((clock() - iStart) / 1000.0);
        printf(" - Moving %d sprites at %2.2f frames per second\n", SDLBenchParams.iNumSprites, fFPS_1);
    }	
    // ------

    // ---Main render loop for iNumSprites2---
    iFrames = 0;	
    iStart = clock();	
    done = 0;	
    while(!done) 
    {				
        ++iFrames;
        MoveSprites(SDLBenchParams.iNumSprites2, iRealSprites);
        if(( (float)(clock() - iStart) / 1000) >= SDLBenchParams.dBaseTime) { done = 1; }
    }
    // Print out some timing information
    if(clock() > iStart) 
    {
        fFPS_2 = (double)iFrames / (double)((clock() - iStart) / 1000.0);
        printf(" - Moving %d sprites at %2.2f frames per second\n\n", SDLBenchParams.iNumSprites2, fFPS_2);
    }	
    // ------
    
    // Integrated mode: extended information output
    if(SDLBenchParams.iIntegrated)
    {
        FILE *fp = fopen("sfml.table","wt");
        fprintf(fp,"          <tr><td align=center><b><font size=2 color=#000080>%s - OpenGL", SDLBENCH_SFML_VERSION);
        fprintf(fp,"</font></b></td>");
        fprintf(fp,"<td align=center><font size=2 color=#800000>%2.2f</font></td>",fFPS_1);
        fprintf(fp,"<td align=center><font size=2 color=#800000>%2.2f</font></td></tr>\n",fFPS_2);
        fclose(fp);
        fp = fopen("SDL-Benchmark.graph","at");
        fprintf(fp,"%s (OpenGL), %2.2f, %2.2f\n", SDLBENCH_SFML_VERSION, fFPS_1, fFPS_2);
        fclose(fp);
    }

    // Close and free memory buffers
    free(positions);
    free(velocities);

    // End!    
    RemoveSprites(iRealSprites);
    delete window;
    return 0;
}

